#@title FINAL (TESE) - Front & Rear Irradiance class { display-mode: "form" }
class Irradiance():
  def __init__(self, url:str="https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/Albedo.csv"):
    self.url = 'https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/Albedo.csv'

  def _get_TMY(self, location, panel_tilt:float, azimuth:float):

    """
    This method gets from PVGIS the necessary data using the TMY.

    Parameters
      ----------
      location
          The location in which the modules will be installed (using the Location() module)
      panel_tilt: float
          The angle in degrees of the modules has with the horizon
      azimuth: float
          The angle the module has with South (South=0º, North = 180º, West = +90º and East = -90º)
      """

    import math
    from datetime import datetime, date, timedelta
    import pandas as pd
    import numpy as np
    from pvmodule import PVGIS

    #params

    #T2m: 2-m air temperature (degree Celsius)
    #RH: relative humidity (%)
    #G(h): Global irradiance on the horizontal plane (W/m2)
    #Gb(n): Beam/direct irradiance on a plane always normal to sun rays (W/m2)
    #Gd(h): Diffuse irradiance on the horizontal plane (W/m2)
    #IR(h): Surface infrared (thermal) irradiance on a horizontal plane (W/m2)
    #WS10m: 10-m total wind speed (m/s)
    #WD10m: 10-m wind direction (0 = N
    #SP: Surface (air) pressure (Pa)

    inputs ,data, metadata = PVGIS().retrieve_tmy(location.latitude,location.longitude)

    data.rename(columns = {'G(h)':'GHI',
                           'Gd(h)':'DHI',
                           'Gb(n)':'DNI',
                           'RH':'Relative Humidity',
                           'T2m':'2m Air Temperature',
                           'WS10m':'10m Wind speed',
                           'WD10m':'10m wind direction',
                           'SP':'Air pressure',
                          }, inplace = True)



    data['Time_H'] = data.index.strftime("%H").astype(float)
    data['Time_M'] = data.index.strftime("%M").astype(float)
    data['Time_S'] = data.index.strftime("%S").astype(float)

    DOY =  pd.DatetimeIndex(data.index.values).day_of_year
    data['Declination'] = (23.45*np.sin(np.deg2rad((360/365)*(284+DOY)))).values

    omega = 0.25*(data['Time_H']*60+data['Time_M']+data['Time_S']/60-12*60)
    data['Hour angle'] = omega

    psi = np.arccos(math.sin(np.deg2rad(location.latitude))*np.sin(np.deg2rad(data['Declination']))+math.cos(np.deg2rad(location.latitude))*np.cos(np.deg2rad(data['Declination']))*np.cos(np.deg2rad(omega)))*180/math.pi

    cos_psi = math.sin(np.deg2rad(location.latitude))*np.sin(np.deg2rad(data['Declination']))+math.cos(np.deg2rad(location.latitude))*np.cos(np.deg2rad(data['Declination']))*np.cos(np.deg2rad(omega))

    Zs = azimuth

    lat_rad = np.deg2rad(location.latitude)
    declination_rad = np.deg2rad(data['Declination'])
    panel_tilt_rad = np.deg2rad(panel_tilt)
    Zs_rad = np.deg2rad(Zs)
    Hour_angle_rad = np.deg2rad(data['Hour angle'])

    cos_phi = np.sin(lat_rad)*np.sin(declination_rad)*np.cos(panel_tilt_rad) - np.cos(lat_rad)*np.sin(declination_rad)*np.sin(panel_tilt_rad)*np.cos(Zs_rad) +\
              np.cos(lat_rad)*np.cos(declination_rad)*np.cos(Hour_angle_rad)*np.cos(panel_tilt_rad) +\
              np.sin(lat_rad)*np.cos(declination_rad)*np.cos(Hour_angle_rad)*np.sin(panel_tilt_rad)*np.cos(Zs_rad) +\
              np.cos(declination_rad)*np.sin(Hour_angle_rad)*np.sin(panel_tilt_rad)*np.sin(Zs_rad)


    data['Rb_front'] = abs(np.where( ((azimuth - 90) <= data['Hour angle']) & (data['Hour angle'] <= (azimuth + 90)), cos_phi/cos_psi, 0))
    data['Rb_rear'] = abs(np.where( (data['Hour angle']< (azimuth - 90)) | (data['Hour angle'] > (azimuth + 90)), cos_phi/cos_psi, 0))



    cols = ['Rb_front', 'Rb_rear']
    data[cols] = data[cols].clip(upper=1)


    data['DOY'] = DOY
    data['Solar Zenith angle'] = psi
    data = data.drop(['Time_H', 'Time_M','Time_S'], axis=1)

    return inputs ,data, metadata











  def irradiance(self, module, location, panel_tilt:float=35, albedo:float=0.2, azimuth:float = 0, Elevation=2, panel_distance:float = None):

    inputs ,data, metadata = Irradiance()._get_TMY(location, panel_tilt, azimuth)


    if panel_distance == None:
      panel_distance = Irradiance()._modules_spacing(module, panel_tilt, data['DOY'], location.latitude)

    GF_beam = Irradiance()._GF_beam(data)
    GF_diffuse =  Irradiance()._GF_diffuse(data, module, panel_distance, panel_tilt)
    GF_reflected = Irradiance()._GF_reflected(data, albedo, module, panel_distance, panel_tilt, azimuth,Elevation)
    G_front = GF_beam + GF_diffuse + GF_reflected
    data['G_Front'] = G_front

    if module['BIPV'] == 'N' :
      data['Total_G'] = G_front
      return inputs ,data, metadata


    GR_beam = Irradiance()._GR_beam(data)
    GR_diffuse =  Irradiance()._GR_diffuse(data, module, panel_distance, panel_tilt)
    GR_reflected = Irradiance()._GR_reflected(data, albedo, module, panel_distance, panel_tilt, azimuth,Elevation)

    G_Rear = GR_beam + GR_diffuse + GR_reflected
    data['G_Rear'] = G_Rear
    data['Total_G'] = G_Rear + G_front

    return inputs ,data, metadata






#
#  FRONT SIDE
#


  def _VF_front_Module2Sky(self, module, panel_distance, panel_tilt):
    import math

    A = float(module['height'])
    H = A
    beta = panel_tilt
    miu = beta
    D = panel_distance

    VF_Module2Sky = H + math.sqrt((A*math.sin(miu*math.pi/180) - H*math.sin(beta*math.pi/180))**2 + (D+H*math.cos(beta*math.pi/180))**2)-math.sqrt((A*math.sin(miu*math.pi/180))**2+D**2)
    VF_Module2Sky = VF_Module2Sky/(2*H)

    return VF_Module2Sky



  def _VF_front_Module2usGround(self, data, module, panel_distance, panel_tilt, azimuth, Elevation):
    import math
    import numpy as np
    import pandas as pd

    A = float(module['height'])
    H = A
    beta = panel_tilt
    miu = beta
    D = panel_distance

    E = Elevation #(Elevação da altura mais baixa do modulo)


    #--- https://arxiv.org/ftp/arxiv/papers/1812/1812.07849.pdf
    gamma = azimuth + 180
    Hour_angle_rad = np.deg2rad(data['Hour angle']) # h em rad
    declination_rad = np.deg2rad(data['Declination']) # delta em rad
    Solar_Zenith_angle_rad = np.deg2rad(data['Solar Zenith angle']) # 0z em rad
    solar_altitude_rad = 90*np.pi/180-np.arcsin(np.cos(Solar_Zenith_angle_rad)) # solar_altitude em rad



    Shadow = np.tan(solar_altitude_rad)*(E + np.sin(beta*np.pi/180)*H) + np.cos(beta*np.pi/180)*H-np.tan(solar_altitude_rad)*E

    S = np.where( (data['GHI'] > 0 )|( data['DHI'] > 0 )| (data['DNI']> 0), Shadow, 0)


    S = pd.DataFrame(S, columns = ['Shadow'] , index=data.index)

    L1 = A * np.cos(miu*np.pi/180) + D - S
    L2 = np.sqrt((A*np.cos(miu*np.pi/180)+D-S+H*np.cos(beta*np.pi/180))**2+(H*np.sin(beta*np.pi/180))**2)

    VF_Module2usGround = ( H + L1 - L2 ) / (2*H)

    return VF_Module2usGround

  def _VF_front_Module2sGround(self, data, module, panel_distance, panel_tilt, azimuth, Elevation):
    import math
    import numpy as np
    import pandas as pd

    A = float(module['height'])
    H = A
    beta = panel_tilt
    miu = beta
    D = panel_distance

    E = Elevation #(Elevação da altura mais baixa do modulo)


    #--- https://arxiv.org/ftp/arxiv/papers/1812/1812.07849.pdf
    gamma = azimuth + 180
    Hour_angle_rad = np.deg2rad(data['Hour angle']) # h em rad
    declination_rad = np.deg2rad(data['Declination']) # delta em rad
    Solar_Zenith_angle_rad = np.deg2rad(data['Solar Zenith angle']) # 0z em rad
    solar_altitude_rad = 90*np.pi/180-np.arcsin(np.cos(Solar_Zenith_angle_rad)) # solar_altitude em rad



    Shadow = np.tan(solar_altitude_rad)*(E + np.sin(beta*np.pi/180)*H) + np.cos(beta*np.pi/180)*H-np.tan(solar_altitude_rad)*E

    S = np.where( (data['GHI'] > 0 )|( data['DHI'] > 0 )| (data['DNI']> 0), Shadow, 0)


    S = pd.DataFrame(S, columns = ['Shadow'], index=data.index)

    L2 = np.sqrt((A*np.cos(miu*np.pi/180)+D-S+H*np.cos(beta*np.pi/180))**2+(H*np.sin(beta*np.pi/180))**2)
    L3 = A*np.cos(miu*np.pi/180)+D
    L4 = np.sqrt((A*np.cos(miu*np.pi/180)+D+H*np.cos(beta*np.pi/180))**2 + (H*np.sin(beta*np.pi/180))**2)
    L5 = A*np.cos(miu*np.pi/180)+D-S

    _VF_Module2sGround = (L2 + L3 - L4 - L5) / (2 * H)

    return _VF_Module2sGround


  def _GF_beam(self, data):

    GF_beam = (data['GHI'] - data['DHI']) * data['Rb_front']

    return GF_beam

  def _GF_diffuse(self, data, module, panel_distance, panel_tilt):

    GF_diffuse = data['DHI']*Irradiance()._VF_front_Module2Sky(module, panel_distance, panel_tilt)

    return GF_diffuse

  def _GF_reflected(self, data, albedo, module, panel_distance, panel_tilt, azimuth, Elevation):

    #VF_Module2sGround - shaded ground

    #VF_Module2usGround - unshaded ground

    VF_Module2Ground = Irradiance()._VF_front_Module2sGround(data, module, panel_distance, panel_tilt,azimuth,Elevation) + Irradiance()._VF_front_Module2usGround(data, module, panel_distance, panel_tilt,azimuth,Elevation)

    GF_reflected = data['GHI']*albedo*VF_Module2Ground['Shadow']

    return GF_reflected


#
# Rear Side
#



  def _VF_rear_Module2Sky(self, module, panel_distance, panel_tilt):
    import math
    import numpy as np

    A = float(module['height'])
    H = A
    beta = panel_tilt
    miu = beta
    D = panel_distance

    L6 = np.sqrt((D+H+np.cos(beta*np.pi/180))**2+(A*np.sin(miu*np.pi/180)-H*np.sin(beta*np.pi/180))**2)
    L7 = np.sqrt((A*np.cos(beta*np.pi/180)+D+H*np.cos(beta*np.pi/180))**2+(H*np.sin(beta*np.pi/180))**2)
    VF_Module2Sky = A + L6 - L7
    VF_Module2Sky = VF_Module2Sky/(2*A)

    return VF_Module2Sky

  def _VF_rear_Module2usGround(self, data, module, panel_distance, panel_tilt, azimuth, Elevation):
    import math
    import numpy as np
    import pandas as pd

    A = float(module['height'])
    H = A
    beta = panel_tilt
    miu = beta
    D = panel_distance

    E = Elevation #(Elevação da altura mais baixa do modulo)


    #--- https://arxiv.org/ftp/arxiv/papers/1812/1812.07849.pdf
    gamma = azimuth + 180
    Hour_angle_rad = np.deg2rad(data['Hour angle']) # h em rad
    declination_rad = np.deg2rad(data['Declination']) # delta em rad
    Solar_Zenith_angle_rad = np.deg2rad(data['Solar Zenith angle']) # 0z em rad
    solar_altitude_rad = 90*np.pi/180-np.arcsin(np.cos(Solar_Zenith_angle_rad)) # solar_altitude em rad



    Shadow = np.tan(solar_altitude_rad)*(E + np.sin(beta*np.pi/180)*H) + np.cos(beta*np.pi/180)*H-np.tan(solar_altitude_rad)*E

    S = np.where( (data['GHI'] > 0 )|( data['DHI'] > 0 )| (data['DNI']> 0), Shadow, 0)

    S = pd.DataFrame(S, columns = ['Shadow'] , index=data.index)

    L8 = np.sqrt((S - A*np.sin(miu*np.pi/180))**2 + (A*np.sin(miu*np.pi/180))**2)
    L3 = A*np.cos(miu*np.pi/180) + D
    L9 = np.sqrt((A*np.sin(miu*np.pi/180))**2+(D)**2)



    VF_Module2usGround = ( L8 + L3 - L9 - S ) / (2*A)


    return VF_Module2usGround

  def _VF_rear_Module2sGround(self, data, module, panel_distance, panel_tilt, azimuth, Elevation):
    import math
    import numpy as np
    import pandas as pd

    A = float(module['height'])
    H = A
    beta = panel_tilt
    miu = beta
    D = panel_distance

    E = Elevation #(Elevação da altura mais baixa do modulo)


    #--- https://arxiv.org/ftp/arxiv/papers/1812/1812.07849.pdf
    gamma = azimuth + 180
    Hour_angle_rad = np.deg2rad(data['Hour angle']) # h em rad
    declination_rad = np.deg2rad(data['Declination']) # delta em rad
    Solar_Zenith_angle_rad = np.deg2rad(data['Solar Zenith angle']) # 0z em rad
    solar_altitude_rad = 90*np.pi/180-np.arcsin(np.cos(Solar_Zenith_angle_rad)) # solar_altitude em rad



    Shadow = np.tan(solar_altitude_rad)*(E + np.sin(beta*np.pi/180)*H) + np.cos(beta*np.pi/180)*H-np.tan(solar_altitude_rad)*E

    S = np.where( (data['GHI'] > 0 )|( data['DHI'] > 0 )| (data['DNI']> 0), Shadow, 0)


    S = pd.DataFrame(S, columns = ['Shadow'], index=data.index)

    L8 = np.sqrt((S - A*np.sin(miu*np.pi/180))**2 + (A*np.sin(miu*np.pi/180))**2)

    _VF_Module2sGround = (A + S - L8) / (2 * A)

    return _VF_Module2sGround





  def _GR_beam(self, data):

    GR_beam = (data['GHI'] - data['DHI']) * data['Rb_rear']

    return GR_beam


  def _GR_diffuse(self, data, module, panel_distance, panel_tilt):

    GR_diffuse = data['DHI']*Irradiance()._VF_rear_Module2Sky(module, panel_distance, panel_tilt)

    return GR_diffuse

  def _GR_reflected(self, data, albedo, module, panel_distance, panel_tilt, azimuth, Elevation):

    VF_rear_Module2usGround = Irradiance()._VF_rear_Module2usGround(data, module, panel_distance, panel_tilt,azimuth,Elevation)

    VF_rear_Module2sGround = Irradiance()._VF_rear_Module2sGround(data, module, panel_distance, panel_tilt,azimuth,Elevation)

    GR_reflected = data['GHI']*albedo*VF_rear_Module2usGround['Shadow'] + data['DHI']*albedo*VF_rear_Module2sGround['Shadow']

    return GR_reflected





    VF_Module2Ground = Irradiance()._VF_front_Module2sGround(data, module, panel_distance, panel_tilt,azimuth,Elevation) + Irradiance()._VF_front_Module2usGround(data, module, panel_distance, panel_tilt,azimuth,Elevation)

    GF_reflected = data['GHI']*albedo*VF_Module2Ground['Shadow']




#
#Auxiliry methods:
#


  def _modules_spacing(self, module, tilt: float, n_year: int = None, latitude: float = None) -> float:
        """
        This method calculates the necessary spacing between modules to eliminate shading/parcial shading.
        To calculate the worst-case scenario, use only:
          module - the module object
          tilt - surface tilt of the module (degree)

        To calculate for a specific time of the year, use in addition
          n_year - the day in the year
          latitude - latitude of the system
        """
        import math
        import numpy as np

        if n_year.empty or latitude == None:
            spacing = round(float(module['height']) * ( math.cos(tilt * math.pi / 180) + (math.sin(tilt * math.pi / 180)) / (math.tan(23.45 * math.pi / 180)) ), 3, )

        else:
            delta = 23.45 * np.sin(((360 / 365) * np.pi / 180) * (n_year+284))
            beta_n = 90 - latitude + delta
            spacing = round( float(module['height']) * ( np.cos(tilt * np.pi / 180) + (np.sin(tilt * np.pi / 180)) / (np.tan(beta_n * np.pi / 180)) ), 3, )

        return max(spacing)

  def list_albedo(self):
    """
    This method returns a list of albedos for specific surfaces.
    """
    import pandas as pd
    from tabulate import tabulate
    albedo_list = pd.read_csv(self.url).replace(" ", "")
    return albedo_list


#location = Location().set_location("Oeiras")
#module = Modules().module('6MN6A270')

#_,data,_ = Irradiance().irradiance(module, location, 35, 0.2)

#display(data.head(24))
#data[['G_Front','G_Rear','Total_G']].max()
#data.to_csv("csv.csv")


#@title PVGIS (TESE) class { display-mode: "form" }
class PVGIS():
    """
    PVGIS class retrieves real-world data from the PVGIS-API.
    It uses by defaults the version v5_2, if wanted, it can be swapped by changing the url with another version.
    The months count start at January=0 and December=11
    """

    def __init__(self):
        # hourly
        self.latitude = None  # latitude
        self.longitude = None  # longitude
        self.usehorizon = None  # usehorizon
        self.userhorizon = None  # userhorizon
        self.raddatabase = None  # raddatabase
        self.startyear = None  # startyear
        self.endyear = None  # endyear
        self.pvcalculation = None  # pvcalculation
        self.peakpower = None  # peakpower
        self.pvtechchoice = None  # pvtechchoice
        self.mountingplace = None  # mountingplace
        self.loss = None  # loss
        self.trackingtype = None  # trackingtype
        self.surface_tilt = None  # surface_tilt
        self.surface_azimuth = None  # surface_azimuth
        self.optimalinclination = None  # optimalinclination
        self.optimalangles = None  # optimalangles
        self.components = None  # components
        self.outputformat = None  # outputformat
        self.url = None  # url
        self.data = None
        # monthly
        self.horirrad = None  # horirrad
        self.optrad = None  # optrad
        self.selectrad = None  # selectrad
        self.angle = None  # angle
        self.mr_dni = None  # mr_dni
        self.d2g = None  # d2g
        self.avtemp = None  # avtemp
        # daily
        self.month = None  # month
        self.angle = None  # angle
        self.aspect = None  # aspect
        self.global_irr = None  # global_irr
        self.glob_2axis = None  # glob_2axis
        self.clearsky = None  # clearsky
        self.clearsky_2axis = None  # clearsky_2axis
        self.showtemperatures = None  # ,showtemperatures
        self.localtime = None  # localtime

    def retrieve_hourly( self, latitude: float, longitude: float, usehorizon: int = 1, userhorizon: int = None, raddatabase: str = None, startyear: int = None, endyear: int = None, pvcalculation: int = 0, peakpower: float = None, pvtechchoice: str = "crystSi", mountingplace: str = "free", loss: float = None, trackingtype: int = 0, surface_tilt: float = 0, surface_azimuth: float = 0, optimalinclination: int = 0, optimalangles: int = 0, components: int = 0, outputformat: str = "json", url: str = "http://re.jrc.ec.europa.eu/api/v5_2/") -> object:
        """
        Hourly Data: This method retrieves real-world data using the PVGIS-API.
        ...
        It outputs 3 dataframes with the following structure:
        Inputs , Outputs, Metedata
        Parameters
        ----------
        latitude: float
          Latitude, in decimal degrees, south is negative.
        longitude: float
          Longitude, in decimal degrees, west is negative.
        usehorizon: int, default = 1,
          Calculate taking into account shadows from high horizon. Value of 1 for "yes".
        userhorizon: int, default = None,
          Height of the horizon at equidistant directions around the point of interest, in degrees. Starting at north and moving clockwise. The series '0,10,20,30,40,15,25,5' would mean the horizon height is 0° due north, 10° for north-east, 20° for east, 30° for south-east, etc.
        raddatabase: str, default = None,
          Name of the radiation database (DB): "PVGIS-SARAH" for Europe, Africa and Asia or "PVGIS-NSRDB" for the Americas between 60°N and 20°S, "PVGIS-ERA5" and "PVGIS-COSMO" for Europe (including high-latitudes), and "PVGIS-CMSAF" for Europe and Africa (will be deprecated). The default DBs are PVGIS-SARAH, PVGIS-NSRDB and PVGIS-ERA5 based on the chosen location.
        startyear: int, default = None,
          First year of the output of hourly averages. Availability varies with the temporal coverage of the radiation DB chosen. The default value is the first year of the DB.
        endyear: int, default = None,
          Final year of the output of hourly averages. Availability varies with the temporal coverage of the radiation DB chosen. The default value is the last year of the DB.
        pvcalculation: int, default = 0,
          If "0" outputs only solar radiation calculations, if "1" outputs the estimation of hourly PV production as well.
        peakpower: float, default = None,
          Nominal power of the PV system, in kW.
        pvtechchoice: str, default = "crystSi",
          PV technology. Choices are: "crystSi", "CIS", "CdTe" and "Unknown".
        mountingplace: str, default = "free",
          Type of mounting of the PV modules. Choices are: "free" for free-standing and "building" for building-integrated.
        loss: float, default = None,
          Sum of system losses, in percent.
        trackingtype: int, default = 0,
          Type of suntracking used, 0=fixed, 1=single horizontal axis aligned north-south, 2=two-axis tracking, 3=vertical axis tracking, 4=single horizontal axis aligned east-west, 5=single inclined axis aligned north-south.
        surface_tilt: float, default = 0,
          Inclination angle from horizontal plane. Not relevant for 2-axis tracking.
        surface_azimuth: float, default = 0,
          Orientation (azimuth) angle of the (fixed) plane, 0=south, 90=west, -90=east. Not relevant for tracking planes.
        optimalinclination: int, default = 0,
          Calculate the optimum inclination angle. Value of 1 for "yes". All other values (or no value) mean "no". Not relevant for 2-axis tracking.
        optimalangles: int, default = 0,
          Calculate the optimum inclination AND orientation angles. Value of 1 for "yes". All other values (or no value) mean "no". Not relevant for tracking planes.
        components: int, default = 0,
          If "1" outputs beam, diffuse and reflected radiation components. Otherwise, it outputs only global values.
        outputformat: str, default = "json",
          Type of output. Choices are: "csv" for the normal csv output with text explanations, "basic" to get only the data output with no text, and "json".
        url: str, default = "http://re.jrc.ec.europa.eu/api/v5_2/",
        """
        import requests
        import pandas as pd

        self.latitude = latitude
        self.longitude = longitude
        self.usehorizon = usehorizon
        self.userhorizon = userhorizon
        self.raddatabase = raddatabase
        self.startyear = startyear
        self.endyear = endyear
        self.pvcalculation = pvcalculation
        self.peakpower = peakpower
        self.pvtechchoice = pvtechchoice
        self.mountingplace = mountingplace
        self.loss = loss
        self.trackingtype = trackingtype
        self.surface_tilt = surface_tilt
        self.surface_azimuth = surface_azimuth
        self.optimalinclination = optimalinclination
        self.optimalangles = optimalangles
        self.components = components
        self.outputformat = outputformat
        self.url = url
        self.data = None

        url = self.url + f"seriescalc?lat={self.latitude}&lon={self.longitude}"

        if self.usehorizon != None:
            url = url + f"&usehorizon={self.usehorizon}"
        if self.userhorizon != None:
            url = url + f"&userhorizon={self.userhorizon}"
        if self.raddatabase != None:
            url = url + f"&raddatabase={self.raddatabase}"
        if self.startyear != None:
            url = url + f"&startyear={self.startyear}"
        if self.endyear != None:
            url = url + f"&endyear={self.endyear}"
        if self.pvcalculation != None:
            url = url + f"&pvcalculation={self.pvcalculation}"
        if self.peakpower != None:
            url = url + f"&peakpower={self.peakpower}"
        if self.pvtechchoice != None:
            url = url + f"&pvtechchoice={self.pvtechchoice}"
        if self.mountingplace != None:
            url = url + f"&mountingplace={self.mountingplace}"
        if self.loss != None:
            url = url + f"&loss={self.loss}"
        if self.trackingtype != None:
            url = url + f"&trackingtype={self.trackingtype}"
        if self.surface_tilt != None:
            url = url + f"&angle={self.surface_tilt}"
        if self.surface_azimuth != None:
            url = url + f"&aspect={self.surface_azimuth}"
        if self.optimalinclination != None:
            url = url + f"&optimalinclination={self.optimalinclination}"
        if self.optimalangles != None:
            url = url + f"&optimalangles={self.optimalangles}"
        if self.components != None:
            url = url + f"&components={self.components}"
        if self.outputformat != None:
            url = url + f"&outputformat={self.outputformat}"

        self.url = url
        data = requests.get(url).json()
        try:
            outputs = data["outputs"]["hourly"]
            outputs = pd.json_normalize(outputs)
            outputs["time"] = pd.to_datetime(outputs.time, format="%Y%m%d:%H%M")

            outputs = outputs.set_index("time")

            inputs = data["inputs"]

            meta = data["meta"]
        except:
            print(f"Error: {data}")
            erro = f"Error: {data}"
            return erro

        self.data = inputs, outputs, meta

        if(outputs['T2m'].mean() == 0 or outputs['WS10m'].mean() == 0):
          if 'v5_2' in self.url:
            return PVGIS().retrieve_hourly(self.latitude, self.longitude, self.usehorizon, self.userhorizon, self.raddatabase, self.startyear, self.endyear, self.pvcalculation, self.peakpower, self.pvtechchoice, self.mountingplace, self.loss, self.trackingtype, self.surface_tilt, self.surface_azimuth, self.optimalinclination, self.optimalangles, self.components, self.outputformat, url = "http://re.jrc.ec.europa.eu/api/v5_1/")

        return self.data

    def retrieve_monthly(self, latitude: float, longitude: float, usehorizon: int = 1, userhorizon: int = None, raddatabase: str = None, startyear: int = None, endyear: int = None, horirrad: int = 1, optrad: int = 0, selectrad: int = 0, angle: int = 0, mr_dni: int = 1, d2g: int = 1, avtemp: int = 1, outputformat: str = "json", url: str = "http://re.jrc.ec.europa.eu/api/v5_2/", ) -> object:
        """
        Monthly Data: This method retrieves real-world data using the PVGIS-API.
        ...
        It outputs 3 dataframes with the following structure:
        Inputs , Outputs, Metadata

        Parameters
        ----------
        latitude: float
          Latitude, in decimal degrees, south is negative.
        longitude: float
          Longitude, in decimal degrees, west is negative.
        usehorizon: int, default = 1,
          Calculate taking into account shadows from high horizon. Value of 1 for "yes".
        userhorizon: int, default = None,
          Height of the horizon at equidistant directions around the point of interest, in degrees. Starting at north and moving clockwise. The series '0,10,20,30,40,15,25,5' would mean the horizon height is 0° due north, 10° for north-east, 20° for east, 30° for south-east, etc.
        raddatabase: str, default = None,
          Name of the radiation database (DB): "PVGIS-SARAH" for Europe, Africa and Asia or "PVGIS-NSRDB" for the Americas between 60°N and 20°S, "PVGIS-ERA5" and "PVGIS-COSMO" for Europe (including high-latitudes), and "PVGIS-CMSAF" for Europe and Africa (will be deprecated). The default DBs are PVGIS-SARAH, PVGIS-NSRDB and PVGIS-ERA5 based on the chosen location (see Figure xx).
        startyear: int, default = None,
          First year of the output of monthly averages. Availability varies with the temporal coverage of the radiation DB chosen. The default value is the first year of the DB.
        endyear: int, default = None,
          Final year of the output of monthly averages. Availability varies with the temporal coverage of the radiation DB chosen. The default value is the last year of the DB.
        horirrad: int, default = 1,
          Output horizontal plane irradiation. Value of 1 for "yes". All other values (or no value) mean "no".
        optrad: int, default = 0,
          Output annual optimal angle plane irradiation. Value of 1 for "yes". All other values (or no value) mean "no".
        selectrad: int, default = 0,
          Output irradiation on plane of selected inclination. Value of 1 for "yes". All other values (or no value) mean "no".
        angle: int, default = 0,
        	Inclination
          Example of the minimum usage:
            n angle for the selected inclination irradiation option.
        mr_dni: int, default = 1,
          Output direct normal irradiation. Value of 1 for "yes". All other values (or no value) mean "no".
        d2g: int, default = 1,
          Output monthly values of the ratio of diffuse to global radiation (horizontal plane). Value of 1 for "yes". All other values (or no value) mean "no".
        avtemp: int, default = 1,
          Output monthly average values of daily (24h) temperature. Value of 1 for "yes". All other values (or no value) mean "no".
        outputformat: st, defaultr = "json",
          Type of output. Choices are: "csv" for the normal csv output with text explanations, "basic" to get only the data output with no text, and "json".
        url: str, default = "http://re.jrc.ec.europa.eu/api/v5_2/",
            PVGIS 5.1: https://re.jrc.ec.europa.eu/api/v5_1/
            PVGIS 5.2: https://re.jrc.ec.europa.eu/api/v5_2/

      """
        import requests
        import pandas as pd

        self.latitude = latitude
        self.longitude = longitude
        self.usehorizon = usehorizon
        self.userhorizon = userhorizon
        self.raddatabase = raddatabase
        self.startyear = startyear
        self.endyear = endyear
        self.horirrad = horirrad
        self.selectrad = selectrad
        self.angle = angle
        self.mr_dni = mr_dni
        self.d2g = d2g
        self.avtemp = avtemp
        self.outputformat = outputformat
        self.url = url
        self.data = None

        url = self.url + f"MRcalc?lat={self.latitude}&lon={self.longitude}"

        if self.usehorizon != None:
            url = url + f"&usehorizon={self.usehorizon}"
        if self.userhorizon != None:
            url = url + f"&userhorizon={self.userhorizon}"
        if self.raddatabase != None:
            url = url + f"&raddatabase={self.raddatabase}"
        if self.startyear != None:
            url = url + f"&startyear={self.startyear}"
        if self.endyear != None:
            url = url + f"&endyear={self.endyear}"
        if self.horirrad != None:
            url = url + f"&horirrad={self.horirrad}"
        if self.selectrad != None:
            url = url + f"&selectrad={self.selectrad}"
        if self.angle != None:
            url = url + f"&angle={self.angle}"
        if self.mr_dni != None:
            url = url + f"&mr_dni={self.mr_dni}"
        if self.avtemp != None:
            url = url + f"&avtemp={self.avtemp}"
        if self.d2g != None:
            url = url + f"&d2g={self.d2g}"
        if self.outputformat != None:
            url = url + f"&outputformat={self.outputformat}"

        self.url = url
        data = requests.get(url).json()
        try:
            outputs = data["outputs"]["monthly"]
            outputs = pd.json_normalize(outputs)

            inputs = data["inputs"]

            meta = data["meta"]
        except:
            print(f"Error: {data}")
            erro = f"Error: {data}"
            return erro

        self.data = inputs, outputs, meta

        return self.data

    def retrieve_daily(self, latitude: float, longitude: float, month: int, usehorizon: int = 1, userhorizon: int = None, raddatabase: str = None, angle: int = 0, aspect: int = 0, global_irr: int = 1, glob_2axis: int = 0, clearsky: int = 0, clearsky_2axis: int = 0, showtemperatures: int = 1, localtime: int = 1, outputformat: str = "json", url: str = "http://re.jrc.ec.europa.eu/api/v5_2/", ) -> object:
        """
        Daily Data: This method retrieves real-world data using the PVGIS-API.
        The months count start at January=0 and December=11
        ...
        It outputs 3 dataframes with the following structure:
        Inputs , Outputs, Metadata

        Parameters
        ----------
        latitude: float,
          Latitude, in decimal degrees, south is negative.
        longitude: float
          Longitude, in decimal degrees, west is negative.
        month: int,
          The value of this parameter should be the number of the month, starting at 1 for January. If you give the value 0 (zero) you will instead get data for all the months.
        usehorizon: int, default = 1,
          Calculate taking into account shadows from high horizon. Value of 1 for "yes".
        userhorizon: int, default = None,
          Height of the horizon at equidistant directions around the point of interest, in degrees. Starting at north and moving clockwise. The series '0,10,20,30,40,15,25,5' would mean the horizon height is 0° due north, 10° for north-east, 20° for east, 30° for south-east, etc.
        raddatabase: str, default = None,
          Name of the radiation database (DB): "PVGIS-SARAH" for Europe, Africa and Asia or "PVGIS-NSRDB" for the Americas between 60°N and 20°S, "PVGIS-ERA5" and "PVGIS-COSMO" for Europe (including high-latitudes), and "PVGIS-CMSAF" for Europe and Africa (will be deprecated). The default DBs are PVGIS-SARAH, PVGIS-NSRDB and PVGIS-ERA5 based on the chosen location (see Figure xx).
        angle: int, default = 0,
          Inclination angle from horizontal plane of the (fixed) PV system.
        aspect: int, default = 0,
          Orientation (azimuth) angle of the (fixed) PV system, 0=south, 90=west, -90=east.
        global_irr: int, default = 1,
          Output the global, direct and diffuse in-plane irradiances. Value of 1 for "yes". All other values (or no value) mean "no".
        glob_2axis: int, default = 0,
          Output the global, direct and diffuse two-axis tracking irradiances. Value of 1 for "yes". All other values (or no value) mean "no".
        clearsky: int, default = 0,
          Output the global clear-sky irradiance. Value of 1 for "yes". All other values (or no value) mean "no".
        clearsky_2axis: in, defaultt = 0,
          Output the global clear-sky two-axis tracking irradiance. Value of 1 for "yes". All other values (or no value) mean "no".
        showtemperatures: int, default = 1,
          Output the daily temperature profile. Value of 1 for "yes". All other values (or no value) mean "no".
        localtime: int, default = 1,
          Output the time in the local time zone (not daylight saving time), instead of UTC. Value of 1 for "yes". All other values (or no value) mean "no".
        outputformat: str, default = "json",
          Type of output. Choices are: "csv" for the normal csv output with text explanations, "basic" to get only the data output with no text, and "json".
        url: str, default = "http://re.jrc.ec.europa.eu/api/v5_2/",
            PVGIS 5.1: https://re.jrc.ec.europa.eu/api/v5_1/
            PVGIS 5.2: https://re.jrc.ec.europa.eu/api/v5_2/

      """
        import requests
        import pandas as pd

        self.month = month
        self.latitude = latitude
        self.longitude = longitude
        self.usehorizon = usehorizon
        self.userhorizon = userhorizon
        self.raddatabase = raddatabase
        self.angle = angle
        self.aspect = aspect
        self.global_irr = global_irr
        self.glob_2axis = glob_2axis
        self.clearsky = clearsky
        self.clearsky_2axis = clearsky_2axis
        self.showtemperatures = showtemperatures
        self.localtime = localtime
        self.outputformat = outputformat
        self.url = url
        self.data = None

        url = (
            self.url
            + f"DRcalc?lat={self.latitude}&lon={self.longitude}&month={self.month}"
        )

        if self.usehorizon != None:
            url = url + f"&usehorizon={self.usehorizon}"
        if self.userhorizon != None:
            url = url + f"&userhorizon={self.userhorizon}"
        if self.raddatabase != None:
            url = url + f"&raddatabase={self.raddatabase}"
        if self.angle != None:
            url = url + f"&angle={self.angle}"
        if self.aspect != None:
            url = url + f"&aspect={self.aspect}"
        if self.global_irr != None:
            url = url + f"&global={self.global_irr}"
        if self.glob_2axis != None:
            url = url + f"&glob_2axis={self.glob_2axis}"
        if self.clearsky != None:
            url = url + f"&clearsky={self.clearsky}"
        if self.clearsky_2axis != None:
            url = url + f"&clearsky_2axis={self.clearsky_2axis}"
        if self.showtemperatures != None:
            url = url + f"&showtemperatures={self.showtemperatures}"
        if self.localtime != None:
            url = url + f"&localtime={self.localtime}"
        if self.outputformat != None:
            url = url + f"&outputformat={self.outputformat}"

        self.url = url
        data = requests.get(url).json()
        try:
            outputs = data["outputs"]["daily_profile"]
            outputs = pd.json_normalize(outputs)
            outputs = outputs.set_index("time")

            inputs = data["inputs"]

            meta = data["meta"]

        except:
            print(f"Error: {data}")
            erro = f"Error: {data}"
            return erro


        input_tmy, output_tmy, metadata_tmy = PVGIS().retrieve_tmy(latitude,longitude)

        output_tmy = output_tmy[output_tmy.index.month == self.month]

        df = pd.DataFrame()
        df = pd.DataFrame(index=outputs.index)
        wind_speed = []

        for hour in range(0,24,1):
          wind_speed.append(output_tmy.iloc[hour:24+hour, :]['WS10m'].mean())

        outputs['WS10m'] = wind_speed

        if(outputs['T2m'].mean() == 0 or outputs['WS10m'].mean() == 0):
          if 'v5_2' in self.url:
            return PVGIS().retrieve_daily(self.latitude, self.longitude, self.month, self.usehorizon, self.userhorizon, self.raddatabase, self.angle, self.aspect, self.global_irr, self.glob_2axis, self.clearsky, self.clearsky_2axis, self.showtemperatures, self.localtime, self.outputformat, url= "http://re.jrc.ec.europa.eu/api/v5_1/", )

        self.data = inputs, outputs, meta
        return self.data

    def retrieve_tmy(self, latitude: float,
                     longitude: float,
                     usehorizon: int = 1,
                     userhorizon: int = None,
                     startyear: int = None,
                     endyear: int = None,
                     outputformat: str = "json",
                     url: str = "http://re.jrc.ec.europa.eu/api/v5_2/", ) -> object:
        """
        Daily Data: This method retrieves real-world data using the PVGIS-API.
        The months count start at January=0 and December=11
        ...
        It outputs 3 dataframes with the following structure:
        Inputs , Outputs, Metadata

        Parameters
        ----------
        latitude: float,
          Latitude, in decimal degrees, south is negative.
        longitude: float
          Longitude, in decimal degrees, west is negative.
        startyear: int, default = None,
          First year of the output of monthly averages. Availability varies with the temporal coverage of the radiation DB chosen. The default value is the first year of the DB.
        endyear: int, default = None,
          Final year of the output of monthly averages. Availability varies with the temporal coverage of the radiation DB chosen. The default value is the last year of the DB.
        usehorizon: int, default = 1,
          Calculate taking into account shadows from high horizon. Value of 1 for "yes".
        userhorizon: int, default = None,
          Height of the horizon at equidistant directions around the point of interest, in degrees. Starting at north and moving clockwise. The series '0,10,20,30,40,15,25,5' would mean the horizon height is 0° due north, 10° for north-east, 20° for east, 30° for south-east, etc.
        outputformat: str, default = "json",
          Type of output. Choices are: "csv" for the normal csv output with text explanations, "basic" to get only the data output with no text, and "json".
        url: str, default = "http://re.jrc.ec.europa.eu/api/v5_2/",
            PVGIS 5.1: https://re.jrc.ec.europa.eu/api/v5_1/
            PVGIS 5.2: https://re.jrc.ec.europa.eu/api/v5_2/

      """
        import requests
        import pandas as pd

        self.latitude = latitude
        self.longitude = longitude
        self.usehorizon = usehorizon
        self.userhorizon = userhorizon
        self.startyear = startyear
        self.endyear = endyear
        self.outputformat = outputformat
        self.url = url
        self.data = None

        url = (
            self.url
            + f"tmy?lat={self.latitude}&lon={self.longitude}"
        )

        if self.usehorizon != None:
            url = url + f"&usehorizon={self.usehorizon}"
        if self.userhorizon != None:
            url = url + f"&userhorizon={self.userhorizon}"
        if self.startyear != None:
            url = url + f"&startyear={self.startyear}"
        if self.endyear != None:
            url = url + f"&endyear={self.endyear}"
        if self.outputformat != None:
            url = url + f"&outputformat={self.outputformat}"


        self.url = url
        data = requests.get(url).json()
        try:
            outputs = data["outputs"]["tmy_hourly"]
            outputs = pd.json_normalize(outputs)
            outputs["time(UTC)"] = pd.to_datetime(outputs["time(UTC)"], format="%Y%m%d:%H%M")
            outputs = outputs.set_index("time(UTC)")
            outputs.index.names = ['time']

            inputs = data["inputs"]

            meta = data["meta"]

        except:
            print(f"Error : {data}")
            return data

        self.data = inputs, outputs, meta
        return self.data



#@title FINAL (TESE) - Module class { display-mode: "form" }
class Modules():
    def __init__(self, url:str='https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/PV_Modules.csv'):
      self.url = 'https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/PV_Modules.csv'


    def list_modules(self,wattage:int = None, BIPV:str = None, print_data = True):
      """
      List of modules.
      Parameters
      ----------
      url : str, default = 'https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/PV_Modules.csv'
          Url to the list of modules. Can also be a .csv file.
      wattage : int, default = None
          Filter modules by a desired Wattage
      BIPV : str, default = None, default values allows both bi-facial and mono-facial modules to appear in the list
          Filter modules by bi-facial or monofacial modules
            Bi-facial = 'Y'
            Mono-facial = 'N'

      """
      import pandas as pd
      modules = pd.read_csv(self.url).replace(" ", "")
      if wattage != None:
        modules = modules.loc[modules['Pmax'] == int(wattage)]

      if BIPV != None:
        modules = modules.loc[modules['BIPV'] == str(BIPV)]


      modules = modules.loc[modules['Short Side'] != str('nan')]

      from tabulate import tabulate
      if print_data:
        print(tabulate(modules, headers='keys', tablefmt='psql'))
      return modules

    def module(self,model:str ,modules_per_string:int=1, number_of_strings:int=1,losses:float=0, first_year_degradation:float=2,annual_degradation:float=0.33, url: str='https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/PV_Modules.csv') -> dict:
      """
      This method defines the module used for the calculations.
      The standarts module is the LG_Neon_2_ LG350N1C-V5 with the following specsheet:
      """

      import pandas as pd
      module = pd.read_csv(self.url).replace(" ", "")

      module = module.loc[module['Model Number'] == model]
      module = module.values.tolist()

      return      {'name': module[0][0]+' '+module[0][1],
                  'height': module[0][19],
                  'length': module[0][18],
                  'pdc': module[0][3],
                  'uoc': module[0][11],
                  'isc': module[0][10],
                  'NOCT': module[0][14],
                  'tc_pmax': module[0][15],
                  'tc_voc': module[0][17],
                  'tc_isc': module[0][16],
                  'modules_per_string': modules_per_string,
                  'number_of_strings': number_of_strings,
                  'losses': losses,
                  'BIPV': module[0][9],
                  'first_year_degradation': first_year_degradation,
                  'annual_degradation': annual_degradation }

    def modules_spacing(self, module, tilt: float, n_year: int = None, latitude: float = None) -> float:
        """
        This method calculates the necessary spacing between modules to eliminate shading/parcial shading.
        To calculate the worst-case scenario, use only:
          module - the module object
          tilt - surface tilt of the module (degree)

        To calculate for a specific time of the year, use in addition
          n_year - the day in the year
          latitude - latitude of the system
        """
        import math

        if n_year == None or latitude == None:
            self.spacing = round(module['height'] * ( math.cos(tilt * math.pi / 180) + (math.sin(tilt * math.pi / 180)) / (math.tan(23.45 * math.pi / 180)) ), 3, )

        else:
            delta = 23.45 * math.sin(((360 / 365) * math.pi / 180) * (n_year - 81))
            beta_n = 90 - latitude + delta
            self.spacing = round( module['height'] * ( math.cos(tilt * math.pi / 180) + (math.sin(tilt * math.pi / 180)) / (math.tan(beta_n * math.pi / 180)) ), 3, )
        return self.spacing




#@title FINAL (TESE) - Location class { display-mode: "form" }
class Location():
    def __init__(self):
        self.city = None  # city
        self.latitude = None  # latitude
        self.longitude = None  # longitude
        self.elevation = None  # elevation
        self.timezone = None  # timezone
        self.name = None  # name

    def set_location(self, city: str= None, latitude: float = None, longitude: float = None, elevation: float = None, timezone: str = None, name: str = None, ) -> object:
        """
        Using the city name, this method will geolocate its coordinates, elevation, timezone.
        To use costum locations, just input the desired parameters and they will overwrite the geolocation.
        Parameters
        ----------
        city: str
          The name of the city in which the system is going to be built.
        latitude: float, default = None,
          A specific latitude to overwrite the automatic search.
        longitude: float, default = None,
          A specific longitude to overwrite the automatic search.
        elevation: float, default = None,
          A specific elevation to overwrite the automatic search.
          This elevation corresponds to how many meters the city is above the sea-level.
        timezone: str, default = None,
          The timezone in which the city is located.
          A specific timezone to overwrite the automatic search.
        name: str, default = None,
          The name of the system. This does not affect anything.
        """

        #from tzwhere import tzwhere
        import requests
        from geopy.geocoders import Nominatim
        import numpy as np
        import warnings

        warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)

        self.city = city
        self.latitude = latitude
        self.longitude = longitude
        self.elevation = elevation
        #self.timezone = timezone
        self.name = name

        if city != None:
          geolocator = Nominatim(user_agent="fabio_almeida_thesis")
          location = geolocator.geocode(str(city))
          if self.latitude == None:
              self.latitude = location.latitude
          if self.longitude == None:
              self.longitude = location.longitude
          #if self.timezone == None:
          #    self.timezone = tzwhere.tzwhere().tzNameAt(location.latitude, location.longitude)
          if self.name == None:
              try:
                self.name = location.address
              except:
                pass
          if self.elevation == None:
              try:
                url = "https://api.opentopodata.org/v1/eudem25m?"
                params = {"locations": f"{location.latitude},{location.longitude}"}
                result = requests.get(url, params)
                if result.ok:
                    self.elevation = result.json()["results"][0]["elevation"]
                else:
                    url = "https://api.open-elevation.com/api/v1/lookup?"
                    params = {"locations": f"{location.latitude},{location.longitude}"}
                    result = requests.get(url, params)
                    self.elevation = result.json()["results"][0]["elevation"]
              except:
                pass
        else:
          geolocator = Nominatim(user_agent="fabio_almeida_thesis")
          latitude_longitude =  str(latitude)+","+  str(longitude)
          location = geolocator.reverse(str(latitude_longitude))
          self.city = str(location.address)
          location = geolocator.geocode(str(location.address))
          if self.latitude == None:
              self.latitude = location.latitude
          if self.longitude == None:
              self.longitude = location.longitude
          #if self.timezone == None:
          #    self.timezone = tzwhere.tzwhere().tzNameAt(location.latitude, location.longitude)
          if self.name == None:
              try:
                self.name = location.address
              except:
                pass
          if self.elevation == None:
              try:
                url = "https://api.opentopodata.org/v1/eudem25m?"
                params = {"locations": f"{location.latitude},{location.longitude}"}
                result = requests.get(url, params)
                if result.ok:
                    self.elevation = result.json()["results"][0]["elevation"]
                else:
                    url = "https://api.open-elevation.com/api/v1/lookup?"
                    params = {"locations": f"{location.latitude},{location.longitude}"}
                    result = requests.get(url, params)
                    self.elevation = result.json()["results"][0]["elevation"]
              except:
                pass
        return self

    def get_info(self, location):

      return {
          'Address':location.name,
          'Latitude':location.latitude,
          'Longitude':location.longitude,
          'Elevation':location.elevation
          #'Timezone':location.timezone
          }


from pvmodule.location import Location
from pvmodule.pvgis import PVGIS
from pvmodule.system import System
from pvmodule.simulation import Simulation
from pvmodule.module import Modules
from pvmodule.inverter import Inverters
from pvmodule.irradiance import Irradiance



#@title FINAL (TESE) Inverters class { display-mode: "form" }
class Inverters():
    def __init__(self, url:str='https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/CEC_Inverters.csv'):
      self.url = 'https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/CEC_Inverters.csv'

    def inverter(self,name):
      """
      Select the inverter from a provided list.
      To access the list, please use the method:
        list_inverters()

      Parameters
      ----------
      name : str
          The Model Number of the inverter, as listed on the list.
      """
      import pandas as pd
      inverters = pd.read_csv(self.url).replace(" ", "")

      return inverters.loc[inverters['Model Number'] == name]

    def list_inverters(self,vac:int=None, pmax:int=None,print_list:bool=False):
      """
      List of inverters provided by CEC.
      Parameters
      ----------
      url : str, default = 'https://raw.githubusercontent.com/fabio-r-almeida/pvmodule/main/CEC_Inverters.csv'
          Url to the list of inverters. Can also be a .csv file.
      vac : str, default = None
        Filters the results that are equal to the AC voltage output

      pmax : int, default = None
        Filters the results that are equal to the Max Power output

      print_list : bool, default = False
        Prints list of inverters

      """
      import pandas as pd
      inverters = pd.read_csv(self.url).replace(" ", "")

      if vac != None:
        inverters = inverters.loc[inverters['Nominal Voltage (Vac)'] == int(vac)]
      if pmax != None:
        inverters = inverters.loc[inverters['Maximum Continuous Output Power (kW)'] == int(pmax)]
      from tabulate import tabulate
      if print_list:
        print(tabulate(inverters, headers='keys', tablefmt='psql'))

      return inverters



    def auto_select_inverter(self,module):
      import pandas as pd

      number_of_modules = module['modules_per_string'] * module['number_of_strings']
      Max_Input_DC_Power = number_of_modules * module['pdc'] / 1000
      Vdcmax = module['modules_per_string'] * module['uoc']
      Idcmax = module['isc']

      inverter_list = Inverters().list_inverters(print_list=False)


      

      inverter_list = inverter_list.loc[inverter_list['Maximum Continuous Output Power (kW)'] >= Max_Input_DC_Power]
      inverter_list = inverter_list.loc[inverter_list['Voltage Maximum (Vdc)'] >= Vdcmax] #MAX MPPT
      inverter_list = inverter_list.loc[inverter_list['Voltage Minimum (Vdc)'] <= Vdcmax] #MIN MPPT
      inverter_list = inverter_list.loc[inverter_list['Max strings input'] <= module['number_of_strings']]
      inverter_list = inverter_list.loc[inverter_list['Maximum Short Circuit Current / String'] >= Idcmax* ] 

      


      inverter = pd.DataFrame(inverter_list)
      inverter['efficiency'] = inverter['Weighted Efficiency (%)']

     
      inverter = inverter.sort_values(by='Maximum Continuous Output Power (kW)', ascending=True)

      if len(inverter) > 0:
        inverter = inverter.drop( inverter.index.to_list()[1:] ,axis = 0 )
        print(inverter)

        return  inverter

      return pd.DataFrame()






from setuptools import setup, find_packages
try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()
VERSION = '0.0.61'
DESCRIPTION = 'A library used to simulate photovoltaic energy production using PVGIS'
# Setting up
setup(
    name="pvmodule",
    version=VERSION,
    author="Fábio Almeida",
    author_email="<fabio-r-almeida@outlook.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=['requests', 'pandas','geopy','numpy','tabulate','datetime'],
    keywords=['Python', 'PVGIS', 'Simulator', 'Photovoltaic', 'Solar energy', 'Solar panels','Solar simulation'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



#@title System class { display-mode: "form" }
class System():
    def __init__(self):
        self.spacing = None
        self.plot_height = None
        self.plot_length = None
        self.total_modules = None
        self.modules_per_string = 1
        self.number_of_strings = 1
        self.name = None
        self.height = None
        self.length = None
        self.width = None
        self.pdc = None
        self.umpp = None
        self.impp = None
        self.uoc = None
        self.isc = None
        self.NOCT = None
        self.tc_pmax = None
        self.tc_voc = None
        self.tc_isc = None
        self.losses = None


    #def __module_quantity(self, MLength, MHeight, PLength, PHeight, tilt):
    #    """
    #    :MLength : Module config Length
    #    :MHeight : Module config Height
    #    :PLength : Plot config Length
    #    :PHeight : Plot config Height
    #    :Inclination: Module Inclination
    #    """
    #    import math
#
    #    modules_in_string = math.floor(PLength / MLength)
    #    modules_spacing = __modules_spacing(MHeight, tilt)
    #    number_of_strings = math.floor(PHeight / modules_spacing)
#
    #    self.total_modules = modules_in_string * number_of_strings
    #    self.modules_per_string = modules_in_string
    #    self.number_of_strings = number_of_strings
#
    #    return self.total_modules, self.modules_per_string, self.number_of_strings

    
        
############################################
# DC Power calculation.

    def _Tcell(self, module, T_amb, Irradiance, wind_speed):
        #Kurtz - https://pdf.sciencedirectassets.com/277910/1-s2.0-S1876610213X00104/1-s2.0-S1876610213016044/main.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjEFkaCXVzLWVhc3QtMSJIMEYCIQC8KgfYrZdUnt7dMO4KJXbgrhHk5QU4ZYGTevVJaz26vQIhAJWcPNCzwINErnCW8XmhYmmi6h87fdKdQ4c6YCxi4ccRKswECEIQBRoMMDU5MDAzNTQ2ODY1IgwnJmllX8y0kAHLrrAqqQQsVDEGiXuqmbwz1YbBbNgCJwQlcRTRJh13IYb8f4k%2BWOAxjyASlwZ0w1kJX4suaWtUHOfXlskEosm09dh2NX1xNysJ3NwlYMPTHQafDK2%2Fg3g7Zc8DIDw9H2TwqI3kN4rx4YkKb72Rqb9GSIbsfDYtubI%2BACL1ajeIr1idCQ%2F2OefyA%2B%2Fw3siLEthT%2BBnJpYDwc4QqhKG8eewdkJwjvJJyQxnPsFfJf72j0wQcUQ0E6xKju1hFZVgQFXFUDP0onXZJr1w6nYmSJp3phHAuseqKoLWHhCuKy0oVUAxZqT9lkEt6wVWrqPQHJQTUfGU8PlBYKAoMZX457IfHgJU1htpOiEmVBQcnP9uQ%2BKO2sMfEE9JmBZwmMD2XnolOJ2p26%2FgUCwXqbN9EwYBQAYo4SYDLn9n24sVavaMrIf9ZqS5HC34LHNe7tvPcEC0J0mHfhzvzUguT93L0mcBq3nHiEgiTVr5JAgYNIPs0Usuj6ObzpIPzK2dWpOnNcAIfns0PayNSpRcNFGiLvW4tCj0Lfq9fxj21aygbHxeOWIXcu32UvghcQMpOAryMJKDcTaWdBlJNJthX2L5n9epY%2BvuioLVtnYjEyIulokBUWUApz8GJaIX6MFv8kh7iGHiDELEpCLghXBU%2Fo%2F7Kqq3dFiXakWrIFHLFiR%2BPbj3iR%2FikYtt1d1NBXpS4Eb2UWCs0oSPigRzmZooPxSQRVYZzqIEdFfFjKkt%2FmjgjXvzfMJ3X85oGOqgBxHDP1OZMxNmHh8neGyQSrtRnFgRB3I0MArd%2BFFPGdNCy76zqcpGb0FSDGNecPpAdFU8V0x%2FrxWrfuYi3RFi0wEplxj11GkRDnrZhmyTzIXA%2BHDAYhorAVUFLWDMRJ4N0Dune077ee5Abs6qxLp%2FjfG61lEslhUvH4tW3GQTI6VgSqbZXjpvzoRcHnDSmVbZV7vKPpqkT3HbPlrGdGf%2FTtoT6M5WOC%2BKk&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20221029T092527Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAQ3PHCVTYW7235T6G%2F20221029%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=9c4646596a060c4f3be2149c2fa429c590dcebc5240246f15d308cdb3384a9cc&hash=802d1c967efca932ba4387d0f917156d45726685192d8dfbc0dcc9187c0ab7aa&host=68042c943591013ac2b2430a89b270f6af2c76d8dfd086a07176afe7c76c2c61&pii=S1876610213016044&tid=spdf-cb1126c0-67cf-43ee-97cb-272f614a6620&sid=8be7f42727776347ad1a55b471ecb6d1ac82gxrqb&type=client&ua=4d545307060057000602&rr=761add2fcb2169b9
        
        #converts the 10m to 2m wind (meter for the height of the panels).
        from math import log, e
        wind_speed_2m = wind_speed * 4.87 /  log(67.8*10 - 5.42)

        Tc_kurtz_2m =  T_amb + Irradiance*e**(-3.473-0.0594*wind_speed_2m)

        return Tc_kurtz_2m, wind_speed_2m




    def _dc_production(self,module, T_amb, Irradiance,Temp,Isolated_module, wind, wind_speed):
      import pandas as pd
      import math

      if wind == False:
        T_cell = T_amb + (module['NOCT'] - 20) / 800 * Irradiance
      else:
        T_cell, wind_speed = System()._Tcell(module, T_amb, Irradiance, wind_speed)
      
      Pdc_max = module['pdc'] * Irradiance / 1000
      delta_T = T_cell - 25
      PV = Pdc_max * (1 + module['tc_pmax'] / 100 * delta_T)
        

      estimated_current = module['isc']*Irradiance/1000

      dc_prod = PV * (module['modules_per_string']*module['number_of_strings'])
      dc_prod = pd.DataFrame(dc_prod)
      dc_prod.columns = ['DC Power']
      dc_prod['G(i)'] = Irradiance
      dc_prod['U (V)'] = PV/estimated_current
      dc_prod['I (A)'] = estimated_current
      dc_prod = dc_prod.fillna(0)
      if Temp == True:
        dc_prod['Cell Temperature'] = T_cell
      if Isolated_module == True:
        dc_prod['1 DC Module'] = PV 
      if wind == True:
        dc_prod['Wind Speed'] = wind_speed 


      return dc_prod

    def _dc_production_losses(self,dc_prod, losses):
      if losses != type(float) or losses != type(int):
        return dc_prod
      else:
        return dc_prod * (1 - losses)


    def dc_production(self, module, T_amb, Irradiance, temp:bool=False, isolated_module:bool=False, dc_losses:bool=False, wind:bool=False,wind_speed=None):
      """
      This method calculates the DC Power a system would output.
      Parameters
      ----------
      module :  DataFrame.dtype
          Photovoltaic module, can be parameterized using the method:
            module()
      T_amb:  DataFrame.series
          Dataframe with the temperatures. Can be retrieved using PVGIS
      Irradiance:
          Dataframe with the Irradiance on the photovoltaic surface. Can be retrieved using PVGIS.
      dc_losses:bool, default=False
          Setting this to true will include losses on the DC side of the system.
          The default value for the losses are zero, has to be changed in the module parameters.
          This value if in percentage
      temp:bool, default=False
          Setting this to True will add an extra column to the output (the temperature column)
      isolated_module:bool, default=False
          Setting this to True will add an extra column to the output, as if the system were to be run only using one module.
          This is not affected by the inverter efficiency. (AC == DC)
      wind:bool, default=False
          Setting this to True will include wind cooling effects on the module.
          wind_speed has to be provided (dataframe)
      wind_speed :  DataFrame.dtype
          Dataframe that includes the wind speed.
      """

      dc_prod = System()._dc_production(module, T_amb, Irradiance,temp ,isolated_module, wind, wind_speed)
      self.dc_prod = dc_prod
      if dc_losses == True:
        dc_prod_losses = System()._dc_production_losses(dc_prod, module['losses'])
        self.dc_prod = dc_prod_losses

      return self.dc_prod

############################################
# AC Power calculation based on a inverter from the CECInverter list provided by CEC.

    def _ac_production_inverter(self, inverter ,v_dc ,p_dc):
      import pandas as pd

      Paco = float(inverter['Paco'])
      Pnt = float(inverter['Pnt'])
      Pso = float(inverter['Pso'])

      power_ac = System()._efficiency(v_dc, p_dc, inverter)      
      power_ac = System()._limits_power(power_ac, p_dc, Paco, Pnt, Pso)

      if isinstance(p_dc, pd.Series):
          power_ac = pd.Series(power_ac, index=p_dc.index)

      return power_ac

    def _efficiency(self, v_dc, p_dc, inverter):
      '''
      Calculate the inverter AC power
      credits: https://www.nrel.gov/docs/fy15osti/64102.pdf
      '''
      Paco = float(inverter['Paco'])
      Pdco = float(inverter['Pdco'])
      Vdco = float(inverter['Vdco'])
      C0 = float(inverter['C0'])
      C1 = float(inverter['C1'])
      C2 = float(inverter['C2'])
      C3 = float(inverter['C3'])
      Pso = float(inverter['Pso'])

      A = Pdco * (1 + C1 * (v_dc - Vdco))
      B = Pso * (1 + C2 * (v_dc - Vdco))
      C = C0 * (1 + C3 * (v_dc - Vdco))

      return (Paco / (A - B) - C * (A - B)) * (p_dc - B) + C * (p_dc - B)**2

    def _limits_power(self, power_ac, p_dc, Paco, Pnt, Pso):
      '''
      Applies maximum power limits to the ac power depending on the selected inverter.
      '''
      import numpy as np

      power_ac = np.minimum(Paco, power_ac)
      min_ac_power = -1.0 * abs(Pnt)
      below_limit = p_dc < Pso

      try:
          power_ac[below_limit] = min_ac_power
      except TypeError:  
          if below_limit:
              power_ac = min_ac_power

      return power_ac







    def ac_production(self, module, T_amb, Irradiance, inverter, ac_losses:float=0, temp:bool=False, isolated_module:bool=False, dc_losses:bool=False, wind:bool=False,wind_speed=None):
      """
      This method calculates the AC Power a system would provide to the grid, using a provided module and a provided inverter. The efficiency for the inverter is calculated using the methos described here: https://www.nrel.gov/docs/fy15osti/64102.pdf
      Parameters
      ----------
      module :  DataFrame.dtype
          Photovoltaic module, can be parameterized using the method:
            module()
      T_amb:  DataFrame.series
          Dataframe with the temperatures. Can be retrieved using PVGIS
      Irradiance:
          Dataframe with the Irradiance on the photovoltaic surface. Can be retrieved using PVGIS
      inverter:
          Dataframe with all the inverter characteristics. Can be selected using:
              list_inverters() -> select_inverters()
      dc_losses:bool, default=False
          Setting this to true will include losses on the DC side of the system.
          The default value for the losses are zero, has to be changed in the module parameters.
          This value if in percentage
      ac_losses: float, default=0
          Setting this to true will include losses on the AD side of the system.
          The default value for the losses are zero. 
          This value if in percentage.
      temp:bool, default=False
          Setting this to True will add an extra column to the output (the temperature column)
      isolated_module:bool, default=False
          Setting this to True will add an extra column to the output, as if the system were to be run only using one module.
          This is not affected by the inverter efficiency. (AC == DC)
      wind:bool, default=False
          Setting this to True will include wind cooling effects on the module.
          wind_speed has to be provided (dataframe)
      wind_speed :  DataFrame.dtype
          Dataframe that includes the wind speed.
      """

      dc_prod = System().dc_production(module=module, T_amb=T_amb, Irradiance = Irradiance,temp=temp, isolated_module=isolated_module ,dc_losses=dc_losses, wind=wind, wind_speed=wind_speed)
      pac_grid_injected = System()._ac_production_inverter(inverter, dc_prod['U (V)'], dc_prod['DC Power']) * (1 - ac_losses)
      pac_grid_injected[pac_grid_injected < 0] = 0
      pac_grid_injected = pac_grid_injected.to_frame()
      pac_grid_injected.columns = ['AC Power']
      pac_grid_injected['DC Power'] = dc_prod['DC Power']
      if wind == True:
        pac_grid_injected['Wind Speed'] = wind_speed
      if temp == True:
        pac_grid_injected['Cell Temperature'] = dc_prod['Cell Temperature']
      if isolated_module == True:
        pac_grid_injected['1 Module AC'] = pac_grid_injected['AC Power'] / (module['modules_per_string']*module['number_of_strings'])
        pac_grid_injected['1 Module DC'] = pac_grid_injected['DC Power'] / (module['modules_per_string']*module['number_of_strings'])

  
      return pac_grid_injected





#@title Simulation class { display-mode: "form" }
class Simulation():
  #def __init__(self):


  def simulate(self,module, location,inverter=None, ac_losses:float=0, dc_losses:bool=False, duration:int = 1, startyear:int = 2005, surface_tilt:float=35, surface_azimuth:float=0, wind:bool=True, download:bool=False, temp:bool=False, isolated_module:bool=False ):
      """
      This method simulates the AC Power and DC Power a system would provide to the grid, using a provided module and a provided inverter (for AC Power). The efficiency for the inverter is calculated using the methos described here: https://www.nrel.gov/docs/fy15osti/64102.pdf
      
      Parameters
      ----------
      location: Location.location
        The location of the desired simulation
      startyear:int , default= 2005
        The startyear of the simulation. When using PVGIS API V5_2, it can go from 2005 to 2020. If wind information if missing, it will use API v5_1 and therefore only go from 2005 to 2016. (this might depend on the location and database data available)
      surface_tilt:float , default=35
        The tilt of the module. This data is being used to fetch real-world data from PVGIS API (degrees)
      surface_azimuth:float , default=35
        The azimuth of the module. This data is being used to fetch real-world data from PVGIS API (degrees)
      duration:int, default= 1
        The duration of the simulation. This will can only be as long as there is available data.
      module :  DataFrame.dtype
          Photovoltaic module, can be parameterized using the method:
            module()
      inverter:
          Dataframe with all the inverter characteristics. Can be selected using:
              list_inverters() -> select_inverters()
      dc_losses:bool, default=False
          Setting this to true will include losses on the DC side of the system.
          The default value for the losses are zero, has to be changed in the module parameters.
          This value if in percentage
      ac_losses: float, default=0
          Setting this to true will include losses on the AD side of the system.
          The default value for the losses are zero. 
          This value if in percentage.
      wind:bool, default=False
          Setting this to True will include wind cooling effects on the module.
          wind_speed has to be provided (dataframe)
      wind_speed :  DataFrame.dtype
          Dataframe that includes the wind speed.
      """
     
      from pvmodule.pvgis import PVGIS
      from pvmodule.system import System
      from pvmodule.inverter import Inverters

      if inverter == None:
        inverter = Inverters().auto_select_inverter(module)
        if len(inverter) <= 0:
          return print('No inverter provided.')

      if startyear < 2005 or startyear > 2020:
        return print('Startyear must be between 2005 and 2020')

      input, output, metadata = PVGIS().retrieve_hourly(latitude=location.latitude, longitude=location.longitude, surface_tilt=surface_tilt, surface_azimuth=surface_azimuth, startyear = startyear)
      output = output.drop(['H_sun','Int'], axis = 1)

      import pandas as pd
      degradation = []
      degradation_year = []

      ac_system = pd.DataFrame()
      dc_system = pd.DataFrame()

      for year in range(duration):
        if startyear + year > 2020:
          break
        if year == 0 :
          output_year = output[output.index.year == startyear + year]
          degradation.append(100)
          degradation_value = 100
          degradation_year.append(startyear + year)
        elif year == 1:
          output_year = output[output.index.year == startyear + year]
          degradation.append(((1-module['first_year_degradation']/100))*100)
          degradation_value = ((1-module['first_year_degradation']/100))*100
          degradation_year.append(startyear + year)
        else: 
          output_year = output[output.index.year == startyear + year]
          degradation.append(((1-module['first_year_degradation']/100)*(1-module['annual_degradation']/100)**(year-1))*100)
          degradation_value = ((1-module['first_year_degradation']/100)*(1-module['annual_degradation']/100)**(year-1))*100
          degradation_year.append(startyear + year)
        for month in range(1,13):
          output_month = output_year[output_year.index.month == month]
          system_ac = System().ac_production(module=module, T_amb=output_month['T2m'], Irradiance=output_month['G(i)'], inverter=inverter, ac_losses=ac_losses, dc_losses=dc_losses, wind=wind, wind_speed=output_month['WS10m'],temp=temp, isolated_module= isolated_module)
          system_ac = system_ac * degradation_value/100        
          ac_system = ac_system.append(system_ac)

        degradation_df = pd.DataFrame(degradation,columns=['Module Degradation'])
        degradation_df['Year'] = degradation_year

        if download == True:
          degradation_df.to_csv(r'degradation.csv')
          ac_system.to_csv(r'ac_system.csv')
          with open("ac_system.csv", "r+") as fp:
            existing=fp.read()
            fp.seek(0) #point to first line
            fp.write(f"Latitude: {location.latitude}\nLongitude: {location.longitude}\nAddress: {location.name.replace(',',' ')}\nTimezone: {location.timezone}\n Elevation: {location.elevation}\n\nAC Power - AC Power generated by the whole system. Limited by the inverters maximum capacity\nDC Power - System's theorical maximum DC Power generated (has no limitations)\n Wind Speed- The wind speed at a height of 2m (m/2)\n Cell Temperature - The cell temperature using the Kurtz formula (in case the wind speed is being considered) \n 1 Module AC / 1 Module DC - The Power as if the system has only 1 module\n\n"+existing) # add a line above the previously exiting first line


      return ac_system , degradation_df

  def simulate_vertical(self,module, location, inverter=None, ac_losses:float=0, dc_losses:bool=False, duration:int = 1, startyear:int = 2005, surface_tilt:float=90, surface_azimuth:float=90, wind:bool=True, download:bool=False, temp:bool=False, isolated_module:bool=False ):
      """
      This method simulates the AC Power and DC Power a system would provide to the grid, using a provided module and a provided inverter (for AC Power). The efficiency for the inverter is calculated using the methos described here: https://www.nrel.gov/docs/fy15osti/64102.pdf
      
      Parameters
      ----------
      location: Location.location
        The location of the desired simulation
      startyear:int , default= 2005
        The startyear of the simulation. When using PVGIS API V5_2, it can go from 2005 to 2020. If wind information if missing, it will use API v5_1 and therefore only go from 2005 to 2016. (this might depend on the location and database data available)
      surface_tilt:float , default=90
        The tilt of the module. This data is being used to fetch real-world data from PVGIS API (degrees)
      surface_azimuth:float , default= 90 and -90
        The azimuth of the module. This data is being used to fetch real-world data from PVGIS API (degrees)
      duration:int, default= 1
        The duration of the simulation. This will can only be as long as there is available data.
      module :  DataFrame.dtype
          Photovoltaic module, can be parameterized using the method:
            module()
      inverter:
          Dataframe with all the inverter characteristics. Can be selected using:
              list_inverters() -> select_inverters()
      dc_losses:bool, default=False
          Setting this to true will include losses on the DC side of the system.
          The default value for the losses are zero, has to be changed in the module parameters.
          This value if in percentage
      ac_losses: float, default=0
          Setting this to true will include losses on the AD side of the system.
          The default value for the losses are zero. 
          This value if in percentage.
      wind:bool, default=False
          Setting this to True will include wind cooling effects on the module.
          wind_speed has to be provided (dataframe)
      wind_speed :  DataFrame.dtype
          Dataframe that includes the wind speed.
      """
     
      from pvmodule.pvgis import PVGIS
      from pvmodule.system import System
      from pvmodule.inverter import Inverters

      if inverter == None:
        inverter = Inverters().auto_select_inverter(module)
        if len(inverter) <= 0:
          return print('No inverter provided.')


      if startyear < 2005 or startyear > 2020:
        return print('Startyear must be between 2005 and 2020')

      input1, output1, metadata1 = PVGIS().retrieve_hourly(latitude=location.latitude, longitude=location.longitude, surface_tilt=surface_tilt, surface_azimuth=surface_azimuth, startyear = startyear)
      output1 = output1.drop(['H_sun','Int'], axis = 1)
      
      input2, output2, metadata2 = PVGIS().retrieve_hourly(latitude=location.latitude, longitude=location.longitude, surface_tilt=surface_tilt, surface_azimuth=(surface_azimuth*-1), startyear = startyear)
      output2 = output2.drop(['H_sun','Int'], axis = 1)

      output = output1.add(output2, fill_value=0)
      output = output.mul({'G(i)': 1, 'T2m': 0.5, 'WS10m':0.5})

      import pandas as pd
      degradation = []
      degradation_year = []

      ac_system = pd.DataFrame()
      dc_system = pd.DataFrame()

      for year in range(duration):
        if startyear + year > 2020:
          break
        if year == 0 :
          output_year = output[output.index.year == startyear + year]
          degradation.append(100)
          degradation_value = 100
          degradation_year.append(startyear + year)
        elif year == 1:
          output_year = output[output.index.year == startyear + year]
          degradation.append(((1-module['first_year_degradation']/100))*100)
          degradation_value = ((1-module['first_year_degradation']/100))*100
          degradation_year.append(startyear + year)
        else: 
          output_year = output[output.index.year == startyear + year]
          degradation.append(((1-module['first_year_degradation']/100)*(1-module['annual_degradation']/100)**(year-1))*100)
          degradation_value = ((1-module['first_year_degradation']/100)*(1-module['annual_degradation']/100)**(year-1))*100
          degradation_year.append(startyear + year)
        for month in range(1,13):
          output_month = output_year[output_year.index.month == month]
          system_ac = System().ac_production(module=module, T_amb=output_month['T2m'], Irradiance=output_month['G(i)'], inverter=inverter, ac_losses=ac_losses, dc_losses=dc_losses, wind=wind, wind_speed=output_month['WS10m'],temp=temp, isolated_module= isolated_module)
          system_ac = system_ac * degradation_value/100        
          ac_system = ac_system.append(system_ac)

        degradation_df = pd.DataFrame(degradation,columns=['Module Degradation'])
        degradation_df['Year'] = degradation_year

        if download == True:
          degradation_df.to_csv(r'degradation.csv')
          ac_system.to_csv(r'ac_system.csv')
          with open("ac_system.csv", "r+") as fp:
            existing=fp.read()
            fp.seek(0) #point to first line
            fp.write(f"Latitude: {location.latitude}\nLongitude: {location.longitude}\nAddress: {location.name.replace(',',' ')}\nTimezone: {location.timezone}\n Elevation: {location.elevation}\n\nAC Power - AC Power generated by the whole system. Limited by the inverters maximum capacity\nDC Power - System's theorical maximum DC Power generated (has no limitations)\n Wind Speed- The wind speed at a height of 2m (m/2)\n Cell Temperature - The cell temperature using the Kurtz formula (in case the wind speed is being considered) \n 1 Module AC / 1 Module DC - The Power as if the system has only 1 module\n\n"+existing) # add a line above the previously exiting first line


      return ac_system , degradation_df